from .version import __version__

import os
motd = open(os.path.dirname(__file__) + '/resource/ascii.txt').read().format(__version__)

from . import scripting
from .lib import material
from .lib.rating import OpenSkillRating
from .overlay import Overlay, OverlayRegistry
from .io import action
from .io.stimulus import Serialized
from .io.action import Action
from .core import config, agent
from .core.agent import Agent
from .core.env import Env
from .systems.achievement import Task
from .core.terrain import MapGenerator, Terrain

__all__ = ['Env', 'config', 'scripting', 'agent', 'Agent', 'MapGenerator', 'Terrain',
        'Serialized', 'action', 'Action', 'scripting', 'material',
        'Task', 'Overlay', 'OverlayRegistry']



import numpy as np
from pdb import set_trace as T

import nmmo
from nmmo.systems import ai, equipment
from nmmo.lib import material

from nmmo.systems.skill import Skills
from nmmo.systems.achievement import Diary
from nmmo.entity import entity

class Player(entity.Entity):
   def __init__(self, realm, pos, agent, color, pop):
      super().__init__(realm, pos, agent.iden, agent.name, color, pop)

      self.agent  = agent
      self.pop    = pop

      #Scripted hooks
      self.target = None
      self.food   = None
      self.water  = None
      self.vision = 7

      #Submodules
      self.skills = Skills(self)

      self.diary  = None
      if tasks := realm.config.TASKS:
          self.diary = Diary(tasks)

      self.dataframe.init(nmmo.Serialized.Entity, self.entID, self.pos)

   @property
   def serial(self):
      return self.population, self.entID

   @property
   def isPlayer(self) -> bool:
      return True

   @property
   def population(self):
      return self.pop

   def applyDamage(self, dmg, style):
      self.resources.food.increment(dmg)
      self.resources.water.increment(dmg)
      self.skills.applyDamage(dmg, style)
      
   def receiveDamage(self, source, dmg):
      if not super().receiveDamage(source, dmg):
         if source:
            source.history.playerKills += 1
         return 

      self.resources.food.decrement(dmg)
      self.resources.water.decrement(dmg)
      self.skills.receiveDamage(dmg)

   def receiveLoot(self, loadout):
      if loadout.chestplate.level > self.loadout.chestplate.level:
         self.loadout.chestplate = loadout.chestplate
      if loadout.platelegs.level > self.loadout.platelegs.level:
         self.loadout.platelegs = loadout.platelegs

   def packet(self):
      data = super().packet()

      data['entID']    = self.entID
      data['annID']    = self.population

      data['base']     = self.base.packet()
      data['resource'] = self.resources.packet()
      data['skills']   = self.skills.packet()

      return data
  
   def update(self, realm, actions):
      '''Post-action update. Do not include history'''
      super().update(realm, actions)

      if not self.alive:
         return

      self.resources.update(realm, self, actions)
      self.skills.update(realm, self, actions)

      if self.diary:
         self.diary.update(realm, self)



from pdb import set_trace as T
import numpy as np

import functools

import gym
from pettingzoo import ParallelEnv

import nmmo
from nmmo import entity, core
from nmmo.core import terrain
from nmmo.lib import log
from nmmo.infrastructure import DataType

class Env(ParallelEnv):
   '''Environment wrapper for Neural MMO using the Parallel PettingZoo API

   Neural MMO provides complex environments featuring structured observations/actions,
   variably sized agent populations, and long time horizons. Usage in conjunction
   with RLlib as demonstrated in the /projekt wrapper is highly recommended.'''

   metadata = {'render.modes': ['human'], 'name': 'neural-mmo'}

   def __init__(self, config=None):
      '''
      Args:
         config : A forge.blade.core.Config object or subclass object
      '''
      super().__init__()

      if config is None:
          config = nmmo.config.Default()

      if __debug__:
         err = 'Config {} is not a config instance (did you pass the class?)'
         assert isinstance(config, nmmo.config.Config), err.format(config)

      if not config.AGENTS:
          from nmmo import agent
          config.AGENTS = [agent.Random]

      if not config.MAP_GENERATOR:
          config.MAP_GENERATOR = terrain.MapGenerator

      self.realm      = core.Realm(config)
      self.registry   = nmmo.OverlayRegistry(config, self)

      self.config     = config
      self.overlay    = None
      self.overlayPos = [256, 256]
      self.client     = None
      self.obs        = None

      self.has_reset  = False

   @functools.lru_cache(maxsize=None)
   def observation_space(self, agent: int):
      '''Neural MMO Observation Space

      Args:
         agent: Agent ID

      Returns:
         observation: gym.spaces object contained the structured observation
         for the specified agent. Each visible object is represented by
         continuous and discrete vectors of attributes. A 2-layer attentional
         encoder can be used to convert this structured observation into
         a flat vector embedding.'''

      observation = {}
      for entity in sorted(nmmo.Serialized.values()):
         rows       = entity.N(self.config)
         continuous = 0
         discrete   = 0

         for _, attr in entity:
            if attr.DISCRETE:
               discrete += 1
            if attr.CONTINUOUS:
               continuous += 1

         name = entity.__name__
         observation[name] = {
               'Continuous': gym.spaces.Box(low=-2**20, high=2**20, shape=(rows, continuous), dtype=DataType.CONTINUOUS),
               'Discrete'  : gym.spaces.Box(low=0, high=4096, shape=(rows, discrete), dtype=DataType.DISCRETE)}

         if name == 'Entity':
            observation['Entity']['N'] = gym.spaces.Box(low=0, high=self.config.N_AGENT_OBS, shape=(1,), dtype=DataType.DISCRETE)

         observation[name] = gym.spaces.Dict(observation[name])

      return gym.spaces.Dict(observation)

   @functools.lru_cache(maxsize=None)
   def action_space(self, agent):
      '''Neural MMO Action Space

      Args:
         agent: Agent ID

      Returns:
         actions: gym.spaces object contained the structured actions
         for the specified agent. Each action is parameterized by a list
         of discrete-valued arguments. These consist of both fixed, k-way
         choices (such as movement direction) and selections from the
         observation space (such as targeting)'''

      actions = {}
      for atn in sorted(nmmo.Action.edges):
         actions[atn] = {}
         for arg in sorted(atn.edges):
            n                       = arg.N(self.config)
            actions[atn][arg] = gym.spaces.Discrete(n)

         actions[atn] = gym.spaces.Dict(actions[atn])

      return gym.spaces.Dict(actions)
 
   ############################################################################
   ### Core API
   def reset(self, idx=None, step=True):
      '''OpenAI Gym API reset function

      Loads a new game map and returns initial observations

      Args:
         idx: Map index to load. Selects a random map by default

         step: Whether to step the environment and return initial obs

      Returns:
         obs: Initial obs if step=True, None otherwise 

      Notes:
         Neural MMO simulates a persistent world. Ideally, you should reset
         the environment only once, upon creation. In practice, this approach
         limits the number of parallel environment simulations to the number
         of CPU cores available. At small and medium hardware scale, we
         therefore recommend the standard approach of resetting after a long
         but finite horizon: ~1000 timesteps for small maps and
         5000+ timesteps for large maps

      Returns:
         observations, as documented by step()
      '''
      self.has_reset = True

      self.actions = {}
      self.dead    = []

      self.quill = log.Quill()
      
      if idx is None:
         idx = np.random.randint(self.config.NMAPS) + 1

      self.worldIdx = idx
      self.realm.reset(idx)

      if step:
         self.obs, _, _, _ = self.step({})

      return self.obs

   def close(self):
       '''For conformity with the PettingZoo API only; rendering is external'''
       pass

   def step(self, actions):
      '''Simulates one game tick or timestep

      Args:
         actions: A dictionary of agent decisions of format::

               {
                  agent_1: {
                     action_1: [arg_1, arg_2],
                     action_2: [...],
                     ...
                  },
                  agent_2: {
                     ...
                  },
                  ...
               }

            Where agent_i is the integer index of the i\'th agent 

            The environment only evaluates provided actions for provided
            agents. Unprovided action types are interpreted as no-ops and
            illegal actions are ignored

            It is also possible to specify invalid combinations of valid
            actions, such as two movements or two attacks. In this case,
            one will be selected arbitrarily from each incompatible sets.

            A well-formed algorithm should do none of the above. We only
            Perform this conditional processing to make batched action
            computation easier.

      Returns:
         (dict, dict, dict, None):

         observations:
            A dictionary of agent observations of format::

               {
                  agent_1: obs_1,
                  agent_2: obs_2,
                  ...
               }

            Where agent_i is the integer index of the i\'th agent and
            obs_i is specified by the observation_space function.
           
         rewards:
            A dictionary of agent rewards of format::

               {
                  agent_1: reward_1,
                  agent_2: reward_2,
                  ...
               }

            Where agent_i is the integer index of the i\'th agent and
            reward_i is the reward of the i\'th' agent.

            By default, agents receive -1 reward for dying and 0 reward for
            all other circumstances. Override Env.reward to specify
            custom reward functions
 
         dones:
            A dictionary of agent done booleans of format::

               {
                  agent_1: done_1,
                  agent_2: done_2,
                  ...
               }

            Where agent_i is the integer index of the i\'th agent and
            done_i is a boolean denoting whether the i\'th agent has died.

            Note that obs_i will be a garbage placeholder if done_i is true.
            This is provided only for conformity with PettingZoo. Your
            algorithm should not attempt to leverage observations outside of
            trajectory bounds. You can omit garbage obs_i values by setting
            omitDead=True.

         infos:
            A dictionary of agent infos of format:

               {
                  agent_1: None,
                  agent_2: None,
                  ...
               }

            Provided for conformity with PettingZoo
      '''
      assert self.has_reset, 'step before reset'

      #Preprocess actions for neural models
      for entID in list(actions.keys()):
         ent = self.realm.players[entID]
         if not ent.alive:
            continue

         self.actions[entID] = {}
         for atn, args in actions[entID].items():
            self.actions[entID][atn] = {}
            for arg, val in args.items():
               if len(arg.edges) > 0:
                  self.actions[entID][atn][arg] = arg.edges[val]
               elif val < len(ent.targets):
                  targ                     = ent.targets[val]
                  self.actions[entID][atn][arg] = self.realm.entity(targ)
               else: #Need to fix -inf in classifier before removing this
                  self.actions[entID][atn][arg] = ent

      #Step: Realm, Observations, Logs
      self.dead    = self.realm.step(self.actions)
      self.actions = {}
      self.obs     = {}
      infos        = {}

      obs, rewards, dones, self.raw = {}, {}, {}, {}
      for entID, ent in self.realm.players.items():
         ob = self.realm.dataframe.get(ent)
         self.obs[entID] = ob
         if ent.agent.scripted:
            atns = ent.agent(ob)
            if nmmo.action.Attack in atns:
               atn  = atns[nmmo.action.Attack]
               targ = atn[nmmo.action.Target]
               atn[nmmo.action.Target] = self.realm.entity(targ)
            self.actions[entID] = atns
         else:
            obs[entID]     = ob
            self.dummy_ob  = ob

            rewards[entID], infos[entID] = self.reward(ent)
            dones[entID]   = False

      for entID, ent in self.dead.items():
         self.log(ent)

      for entID, ent in self.dead.items():
         if ent.agent.scripted:
            continue
         rewards[ent.entID], infos[ent.entID] = self.reward(ent)
         dones[ent.entID]   = True
         obs[ent.entID]     = self.dummy_ob

      #Pettingzoo API
      self.agents = list(self.realm.players.keys())

      self.obs = obs
      return obs, rewards, dones, infos

   ############################################################################
   ### Logging
   def log(self, ent) -> None:
      '''Logs agent data upon death

      This function is called automatically when an agent dies. Logs are used
      to compute summary stats and populate the dashboard. You should not
      call it manually. Instead, override this method to customize logging.

      Args:
         ent: An agent
      '''

      quill = self.quill

      blob = quill.register('Population', self.realm.tick)
      blob.log(self.realm.population)

      blob = quill.register('Lifetime', self.realm.tick)
      blob.log(ent.history.timeAlive.val)

      blob = quill.register('Skill Level', self.realm.tick)
      blob.log(ent.skills.range.level,        'Range')
      blob.log(ent.skills.mage.level,         'Mage')
      blob.log(ent.skills.melee.level,        'Melee')
      blob.log(ent.skills.constitution.level, 'Constitution')
      blob.log(ent.skills.defense.level,      'Defense')
      blob.log(ent.skills.fishing.level,      'Fishing')
      blob.log(ent.skills.hunting.level,      'Hunting')

      blob = quill.register('Equipment', self.realm.tick)
      blob.log(ent.loadout.chestplate.level, 'Chestplate')
      blob.log(ent.loadout.platelegs.level,  'Platelegs')

      blob = quill.register('Exploration', self.realm.tick)
      blob.log(ent.history.exploration)

      quill.stat('Lifetime',  ent.history.timeAlive.val)

      if ent.diary:
         quill.stat('Tasks_Completed', ent.diary.completed)
         quill.stat('Task_Reward', ent.diary.cumulative_reward)
         for achievement in ent.diary.achievements:
            quill.stat(achievement.name, float(achievement.completed))
      else:
         quill.stat('Task_Reward', ent.history.timeAlive.val)

      quill.stat('PolicyID', ent.agent.policyID)

   def terminal(self):
      '''Logs currently alive agents and returns all collected logs

      Automatic log calls occur only when agents die. To evaluate agent
      performance over a fixed horizon, you will need to include logs for
      agents that are still alive at the end of that horizon. This function
      performs that logging and returns the associated a data structure
      containing logs for the entire evaluation

      Args:
         ent: An agent

      Returns:
         Log datastructure
      '''

      for entID, ent in self.realm.players.entities.items():
         self.log(ent)

      return self.quill.packet

   ############################################################################
   ### Override hooks
   def reward(self, player):
      '''Computes the reward for the specified agent

      Override this method to create custom reward functions. You have full
      access to the environment state via self.realm. Our baselines do not
      modify this method; specify any changes when comparing to baselines

      Args:
         player: player object

      Returns:
         reward:
            The reward for the actions on the previous timestep of the
            entity identified by entID.
      '''
      info = {'population': player.pop}
 
      if player.entID not in self.realm.players:
         return -1, info

      if not player.diary:
         return 0, info

      achievement_rewards = player.diary.update(self.realm, player)
      reward = sum(achievement_rewards.values())

      info = {**info, **achievement_rewards}
      return reward, info
      

   ############################################################################
   ### Client data
   def render(self, mode='human') -> None:
      '''Data packet used by the renderer

      Returns:
         packet: A packet of data for the client
      '''

      assert self.has_reset, 'render before reset'

      packet = {
            'config': self.config,
            'pos': self.overlayPos,
            'wilderness': 0
            }

      packet = {**self.realm.packet(), **packet}

      if self.overlay is not None:
         print('Overlay data: ', len(self.overlay))
         packet['overlay'] = self.overlay
         self.overlay      = None

      if not self.client:
         from nmmo.websocket import Application
         self.client = Application(self) 

      pos, cmd = self.client.update(packet)
      self.registry.step(self.obs, pos, cmd)

   def register(self, overlay) -> None:
      '''Register an overlay to be sent to the client

      The intended use of this function is: User types overlay ->
      client sends cmd to server -> server computes overlay update -> 
      register(overlay) -> overlay is sent to client -> overlay rendered

      Args:
         values: A map-sized (self.size) array of floating point values
      '''
      err = 'overlay must be a numpy array of dimension (*(env.size), 3)'
      assert type(overlay) == np.ndarray, err
      self.overlay = overlay.tolist()

   def dense(self):
      '''Simulates an agent on every tile and returns observations

      This method is used to compute per-tile visualizations across the
      entire map simultaneously. To do so, we spawn agents on each tile
      one at a time. We compute the observation for each agent, delete that
      agent, and go on to the next one. In this fashion, each agent receives
      an observation where it is the only agent alive. This allows us to
      isolate potential influences from observations of nearby agents

      This function is slow, and anything you do with it is probably slower.
      As a concrete example, consider that we would like to visualize a
      learned agent value function for the entire map. This would require
      computing a forward pass for one agent per tile. To cut down on
      computation costs, we omit lava tiles from this method

      Returns:
         (dict, dict):

         observations:
            A dictionary of agent observations as specified by step()

         ents:
            A corresponding dictionary of agents keyed by their entID
      '''
      config  = self.config
      R, C    = self.realm.map.tiles.shape

      entID   = 100000
      pop     = 0
      name    = "Value"
      color   = (255, 255, 255)


      observations, ents = {}, {}
      for r in range(R):
         for c in range(C):
            tile    = self.realm.map.tiles[r, c]
            if not tile.habitable:
               continue

            current = tile.ents
            n       = len(current)
            if n == 0:
               ent = entity.Player(self.realm, (r, c), entID, pop, name, color)
            else:
               ent = list(current.values())[0]

            obs = self.realm.dataframe.get(ent)
            if n == 0:
               self.realm.dataframe.remove(nmmo.Serialized.Entity, entID, ent.pos)

            observations[entID] = obs
            ents[entID] = ent
            entID += 1

      return observations, ents



from typing import List

#from forge.blade.core import material
from nmmo.systems import ai

import math

import numpy as np


def map_to_rewards(tiles, entity) -> List[List[float]]:
   lava_reward = stone_reward = water_reward = float('-inf')
   forest_reward = 1.0 + math.pow(
      (1 - entity.resources.food.val / entity.resources.food.max) * 15.0,
      1.25)
   scrub_reward = 1.0
   around_water_reward = 1.0 + math.pow(
      (1 - entity.resources.water.val / entity.resources.water.max) * 15.0,
      1.25)

   reward_matrix = np.full((len(tiles), len(tiles[0])), 0.0)

   for line in range(len(tiles)):
      tile_line = tiles[line]
      for column in range(len(tile_line)):
         tile_val = tile_line[column].state.tex
         if tile_val == 'lava':
            reward_matrix[line][column] += lava_reward

         if tile_val == 'stone':
            reward_matrix[line][column] += stone_reward

         if tile_val == 'forest':
            reward_matrix[line][column] += forest_reward

         if tile_val == 'water':
            reward_matrix[line][column] += water_reward

         #TODO: Make these comparisons work off of the water Enum type
         #instead of string compare
         if 'water' in ai.utils.adjacentMats(tiles, (line, column)):
            reward_matrix[line][column] += around_water_reward

         if tile_val == 'scrub':
            reward_matrix[line][column] += scrub_reward

   return reward_matrix


def compute_values(reward_matrix: List[List[float]]) -> List[List[float]]:
   gamma_factor = 0.8  # look ahead ∈ [0, 1]
   max_delta = 0.01  # maximum allowed approximation

   value_matrix = np.full((len(reward_matrix), len(reward_matrix[0])), 0.0)

   delta = float('inf')
   while delta > max_delta:
      old_value_matrix = np.copy(value_matrix)
      for line in range(len(reward_matrix)):
         for column in range(len(reward_matrix[0])):
            reward = reward_matrix[line][column]
            value_matrix[line][
               column] = reward + gamma_factor * max_value_around(
               (line, column), value_matrix)

      delta = np.amax(
         np.abs(np.subtract(old_value_matrix, value_matrix)))
   return value_matrix


def values_around(position: (int, int), value_matrix: List[List[float]]) -> (
        float, float, float, float):
   line, column = position

   if line - 1 >= 0:
      top_value = value_matrix[line - 1][column]
   else:
      top_value = float('-inf')

   if line + 1 < len(value_matrix):
      bottom_value = value_matrix[line + 1][column]
   else:
      bottom_value = float('-inf')

   if column - 1 >= 0:
      left_value = value_matrix[line][column - 1]
   else:
      left_value = float('-inf')

   if column + 1 < len(value_matrix[0]):
      right_value = value_matrix[line][column + 1]
   else:
      right_value = float('-inf')

   return top_value, bottom_value, left_value, right_value


def max_value_around(position: (int, int),
                     value_matrix: List[List[float]]) -> float:
   return max(values_around(position, value_matrix))


def max_value_position_around(position: (int, int),
                              value_matrix: List[List[float]]) -> (int, int):
   line, column = position
   top_value, bottom_value, left_value, right_value = values_around(position,
                                                                    value_matrix)

   max_value = max(top_value, bottom_value, left_value, right_value)

   if max_value == top_value:
      return line - 1, column
   elif max_value == bottom_value:
      return line + 1, column
   elif max_value == left_value:
      return line, column - 1
   elif max_value == right_value:
      return line, column + 1


def max_value_direction_around(position: (int, int),
                               value_matrix: List[List[float]]) -> (int, int):
   top_value, bottom_value, left_value, right_value = values_around(position,
                                                                    value_matrix)

   max_value = max(top_value, bottom_value, left_value, right_value)

   if max_value == top_value:
      return -1, 0
   elif max_value == bottom_value:
      return 1, 0
   elif max_value == left_value:
      return 0, -1
   elif max_value == right_value:
      return 0, 1



from pdb import set_trace as T
import abc

import numpy as np
from nmmo.systems import experience, combat, ai

from nmmo.lib import material

### Infrastructure ###
class SkillGroup:
   def __init__(self, realm):
      self.expCalc = experience.ExperienceCalculator()
      self.config  = realm.dataframe.config
      self.skills  = set()

   def update(self, realm, entity, actions):
      for skill in self.skills:
         skill.update(realm, entity)

   def packet(self):
      data = {}
      for skill in self.skills:
         data[skill.__class__.__name__.lower()] = skill.packet()

      return data

class Skill:
   skillItems = abc.ABCMeta

   def __init__(self, skillGroup):
      self.config  = skillGroup.config
      self.expCalc = skillGroup.expCalc
      self.exp     = 0

      skillGroup.skills.add(self)

   def packet(self):
      data = {}

      data['exp']   = self.exp
      data['level'] = self.level

      return data

   def update(self, realm, entity):
      pass

   def setExpByLevel(self, level):
      self.exp = self.expCalc.expAtLevel(level)

   @property
   def level(self):
      lvl = self.expCalc.levelAtExp(self.exp)
      assert lvl == int(lvl)
      return int(lvl)

### Skill Subsets ###
class Harvesting(SkillGroup):
   def __init__(self, realm):
      super().__init__(realm)

      self.fishing      = Fishing(self)
      self.hunting      = Hunting(self)

class Combat(SkillGroup):
   def __init__(self, realm):
      super().__init__(realm)

      self.constitution = Constitution(self)
      self.defense      = Defense(self)
      self.melee        = Melee(self)
      self.range        = Range(self)
      self.mage         = Mage(self)

   def packet(self):
      data          = super().packet() 
      data['level'] = combat.level(self)

      return data

   def applyDamage(self, dmg, style):
      if not self.config.game_system_enabled('Progression'):
         return

      config    = self.config
      baseScale = config.PROGRESSION_BASE_XP_SCALE
      combScale = config.PROGRESSION_COMBAT_XP_SCALE
      conScale  = config.PROGRESSION_CONSTITUTION_XP_SCALE

      self.constitution.exp += dmg * baseScale * conScale

      skill      = self.__dict__[style]
      skill.exp += dmg * baseScale * combScale

   def receiveDamage(self, dmg):
      if not self.config.game_system_enabled('Progression'):
         return

      config    = self.config
      baseScale = config.PROGRESSION_BASE_XP_SCALE
      combScale = config.PROGRESSION_COMBAT_XP_SCALE
      conScale  = config.PROGRESSION_CONSTITUTION_XP_SCALE

      self.constitution.exp += dmg * baseScale * conScale
      self.defense.exp      += dmg * baseScale * combScale

class Skills(Harvesting, Combat):
   pass

### Individual Skills ###
class CombatSkill(Skill): pass

class Constitution(CombatSkill):
   def __init__(self, skillGroup):
      super().__init__(skillGroup)
      self.setExpByLevel(self.config.BASE_HEALTH)

   def update(self, realm, entity):
      health = entity.resources.health
      food   = entity.resources.food
      water  = entity.resources.water
      config = self.config

      if not config.game_system_enabled('Resource'):
         health.increment(1)
         return

      # Heal if above fractional resource threshold
      regen       = config.RESOURCE_HEALTH_REGEN_THRESHOLD
      foodThresh  = food > regen * entity.skills.hunting.level
      waterThresh = water > regen * entity.skills.fishing.level

      if foodThresh and waterThresh:
         restore = config.RESOURCE_HEALTH_RESTORE_FRACTION
         restore = np.floor(restore * self.level)
         health.increment(restore)

      if food.empty:
         health.decrement(1)

      if water.empty:
         health.decrement(1)

class Melee(CombatSkill): pass
class Range(CombatSkill): pass
class Mage(CombatSkill): pass
class Defense(CombatSkill): pass

class Fishing(Skill):
   def __init__(self, skillGroup):
      super().__init__(skillGroup)
      config, level = self.config, 1
      if config.game_system_enabled('Progression'):
         level = config.PROGRESSION_BASE_RESOURCE
      elif config.game_system_enabled('Resource'):
         level = config.RESOURCE_BASE_RESOURCE

      self.setExpByLevel(level)

   def update(self, realm, entity):
      if not self.config.game_system_enabled('Resource'):
         return

      water = entity.resources.water
      water.decrement(1)

      if material.Water not in ai.utils.adjacentMats(
            realm.map.tiles, entity.pos):
         return

      restore = self.config.RESOURCE_HARVEST_RESTORE_FRACTION
      restore = np.floor(restore * self.level)
      water.increment(restore)

      if self.config.game_system_enabled('Progression'):
         self.exp += self.config.PROGRESSION_BASE_XP_SCALE * restore

class Hunting(Skill):
   def __init__(self, skillGroup):
      super().__init__(skillGroup)
      config, level = self.config, 1
      if config.game_system_enabled('Progression'):
         level = config.PROGRESSION_BASE_RESOURCE
      elif config.game_system_enabled('Resource'):
         level = config.RESOURCE_BASE_RESOURCE

      self.setExpByLevel(level)

   def update(self, realm, entity):
      if not self.config.game_system_enabled('Resource'):
         return

      food = entity.resources.food
      food.decrement(1)

      r, c = entity.pos
      if (type(realm.map.tiles[r, c].mat) not in [material.Forest] or
            not realm.map.harvest(r, c)):
         return

      restore = self.config.RESOURCE_HARVEST_RESTORE_FRACTION
      restore = np.floor(restore * self.level)
      food.increment(restore)

      if self.config.game_system_enabled('Progression'):
         self.exp += self.config.PROGRESSION_BASE_XP_SCALE * restore



from pdb import set_trace as T
import numpy as np

from collections import defaultdict, deque
import inspect

class staticproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()

class classproperty(object):
    def __init__(self, f):
        self.f = f
    def __get__(self, obj, owner):
        return self.f(owner)

class Iterable(type):
   def __iter__(cls):
      queue = deque(cls.__dict__.items())
      while len(queue) > 0:
         name, attr = queue.popleft()
         if type(name) != tuple:
            name = tuple([name])
         if not inspect.isclass(attr):
            continue
         yield name, attr

   def values(cls):
      return [e[1] for e in cls]

class StaticIterable(type):
    def __iter__(cls):
        stack = list(cls.__dict__.items())
        for name, attr in cls.__dict__.items():
            if name == '__module__':
                continue
            if name.startswith('__'):
                break
            yield name, attr

class NameComparable(type):
   def __hash__(self):
      return hash(self.__name__)

   def __eq__(self, other):
      try:
         return self.__name__ == other.__name__
      except:
         print('Some sphinx bug makes this block doc calls. You should not see this in normal NMMO usage')

   def __ne__(self, other):
      return self.__name__ != other.__name__

   def __lt__(self, other):
      return self.__name__ < other.__name__

   def __le__(self, other):
      return self.__name__ <= other.__name__

   def __gt__(self, other):
      return self.__name__ > other.__name__

   def __ge__(self, other):
      return self.__name__ >= other.__name__

class IterableNameComparable(Iterable, NameComparable):
   pass

def seed():
   return int(np.random.randint(0, 2**32))

def linf(pos1, pos2):
   r1, c1 = pos1
   r2, c2 = pos2
   return max(abs(r1 - r2), abs(c1 - c2))

#Bounds checker
def inBounds(r, c, shape, border=0):
   R, C = shape
   return (
         r > border and
         c > border and
         r < R - border and
         c < C - border
         )




from pdb import set_trace as T
import numpy as np

from signal import signal, SIGINT
import sys, os, json, pickle, time
import threading
import ray

from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.python import log
from twisted.web.server import Site
from twisted.web.static import File

from autobahn.twisted.websocket import WebSocketServerFactory, \
    WebSocketServerProtocol
from autobahn.twisted.resource import WebSocketResource

class GodswordServerProtocol(WebSocketServerProtocol):
    def __init__(self):
        super().__init__()
        print("Created a server")
        self.frame = 0

        #"connected" is already used by WSSP
        self.sent_environment = False
        self.isConnected      = False

        self.pos = [0, 0]
        self.cmd = None

    def onOpen(self):
        print("Opened connection to server")

    def onClose(self, wasClean, code=None, reason=None):
        self.isConnected = False
        print('Connection closed')

    def connectionMade(self):
        super().connectionMade()
        self.factory.clientConnectionMade(self)

    def connectionLost(self, reason):
        super().connectionLost(reason)
        self.factory.clientConnectionLost(self)
        self.sent_environment = False

    #Not used without player interaction
    def onMessage(self, packet, isBinary):
        print("Server packet", packet)
        packet    = packet.decode()
        _, packet = packet.split(';') #Strip headeer
        r, c, cmd = packet.split(' ') #Split camera coords
        if len(cmd) == 0 or cmd == '\t':
            cmd = None

        self.pos = [int(r), int(c)]
        self.cmd = cmd

        self.isConnected = True

    def onConnect(self, request):
        print("WebSocket connection request: {}".format(request))
        realm = self.factory.realm
        self.realm = realm
        self.frame += 1

    def serverPacket(self):
        data = self.realm.packet
        return data

    def sendUpdate(self, data):
        packet               = {}
        packet['resource']   = data['resource']
        packet['player']     = data['player']
        packet['npc']        = data['npc']
        packet['pos']        = data['pos']
        packet['wilderness'] = data['wilderness']

        config = data['config']

        print('Is Connected? : {}'.format(self.isConnected))
        if not self.sent_environment:
            packet['map']    = data['environment']
            packet['border'] = config.TERRAIN_BORDER
            packet['size']   = config.TERRAIN_SIZE

        if 'overlay' in data:
           packet['overlay'] = data['overlay']
           print('SENDING OVERLAY: ', len(packet['overlay']))

        packet = json.dumps(packet).encode('utf8')
        self.sendMessage(packet, False)

class WSServerFactory(WebSocketServerFactory):
    def __init__(self, ip, realm):
        super().__init__(ip)
        self.realm = realm
        self.time = time.time()
        self.clients = []

        self.pos = [0, 0]
        self.cmd = None
        self.tickRate = 0.6
        self.tick = 0

    def update(self, packet):
        self.tick += 1
        uptime = np.round(self.tickRate*self.tick, 1)
        delta = time.time() - self.time
        print('Wall Clock: ', str(delta)[:5], 'Uptime: ', uptime, ', Tick: ', self.tick)
        delta = self.tickRate - delta    
        if delta > 0:
           time.sleep(delta)
        self.time = time.time()

        for client in self.clients:
            client.sendUpdate(packet)
            if client.pos is not None:
                self.pos = client.pos
                self.cmd = client.cmd

        return self.pos, self.cmd

    def clientConnectionMade(self, client):
        self.clients.append(client)

    def clientConnectionLost(self, client):
        self.clients.remove(client)

class Application:
   def __init__(self, realm):
      signal(SIGINT, self.kill)
      log.startLogging(sys.stdout)

      port = 8080
      self.factory          = WSServerFactory(u'ws://localhost:{}'.format(port), realm)
      self.factory.protocol = GodswordServerProtocol 
      resource              = WebSocketResource(self.factory)

      root = File(".")
      root.putChild(b"ws", resource)
      site = Site(root)

      reactor.listenTCP(port, site)

      def run():
          reactor.run(installSignalHandlers=0)

      threading.Thread(target=run).start()

   def update(self, packet):
      return self.factory.update(packet)

   def kill(*args):
      print("Killed by user")
      reactor.stop()
      os._exit(0)



from pdb import set_trace as T

from collections import defaultdict
import numpy as np

import openskill


def rank(policy_ids, scores):
    '''Compute policy rankings from per-agent scores'''
    agents = defaultdict(list)
    for policy_id, score in zip(policy_ids, scores):
        agents[policy_id].append(score + 1e-8*np.random.normal())

    # Double argsort returns ranks
    return np.argsort(np.argsort(
        [-np.mean(vals) for policy, vals in 
        sorted(agents.items())])).tolist()


class OpenSkillRating:
    '''OpenSkill Rating wrapper for estimating relative policy skill

    Provides a simple method for updating skill estimates from raw
    per-agent scores as are typically returned by the environment.'''
    def __init__(self, agents, anchor, mu=1000, sigma=100/3, anchor_mu=1500):
        '''
        Args:
            agents: List of agent classes to rank
            anchor: Baseline policy name to anchor to mu
            mu: Anchor point for the baseline policy (cannot be exactly 0)
            sigma: 68/95/99.7 win rate against 1/2/3 sigma lower SR'''

        if __debug__:
            err = 'Agents must be ordered (e.g. list, not set)'
            assert type(agents) != set, err

        self.ratings = {agent:
                openskill.Rating(mu=mu, sigma=sigma)
                for agent in agents}

        self.mu        = mu
        self.sigma     = sigma
        self.anchor    = anchor
        self.anchor_mu = anchor_mu

        self.anchor_baseline()

    def update(self, ranks=None, policy_ids=None, scores=None):
        '''Updates internal skill rating estimates for each policy

        You should call this function once per simulated environment
        Provide either ranks OR policy_ids and scores

        Args:
            ranks: List of ranks in the same order as agents
            policy_ids: List of policy IDs for each agent episode
            scores: List of scores for each agent episode

        Returns:
            Dictionary of ratings keyed by agent names'''

        if __debug__:
            err = 'Specify either ranks or policy_ids and scores'
            assert ranks is None  != (policy_ids is None and scores is None), err

        if ranks is None:
            ranks = rank(policy_ids, scores)

        teams = [[e] for e in list(self.ratings.values())]
        ratings = openskill.rate(teams, rank=ranks)
        ratings = [openskill.create_rating(team[0]) for team in ratings]
        for agent, rating in zip(self.ratings, ratings):
            self.ratings[agent] = rating

        self.anchor_baseline()

        return self.ratings

    def anchor_baseline(self):
        '''Resets the anchor point policy to mu SR'''
        for agent, rating in self.ratings.items():
            rating.sigma = self.sigma
            if agent == self.anchor:
                rating.mu    = self.anchor_mu
                rating.sigma = self.sigma



from pdb import set_trace as T
import numpy as np
import os

import nmmo
from nmmo.lib import utils

class SequentialLoader:
    '''config.AGENT_LOADER that spreads out agent populations'''
    def __init__(self, config):
        items = config.AGENTS
        for idx, itm in enumerate(items):
           itm.policyID = idx 

        self.items = items
        self.idx   = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.idx = (self.idx + 1) % len(self.items)
        return self.idx, self.items[self.idx]

class TeamLoader:
    '''config.AGENT_LOADER that loads agent populations adjacent'''
    def __init__(self, config):
        items = config.AGENTS
        self.team_size = config.NENT // config.NPOP

        for idx, itm in enumerate(items):
           itm.policyID = idx 

        self.items = items
        self.idx   = -1

    def __iter__(self):
        return self

    def __next__(self):
        self.idx += 1
        team_idx  = self.idx // self.team_size
        return team_idx, self.items[team_idx]


class Template(metaclass=utils.StaticIterable):
   def __init__(self):
      self.data = {}
      cls       = type(self)

      #Set defaults from static properties
      for k, v in cls:
         self.set(k, v)

   def override(self, **kwargs):
      for k, v in kwargs.items():
         err = 'CLI argument: {} is not a Config property'.format(k)
         assert hasattr(self, k), err
         self.set(k, v)

   def set(self, k, v):
      if type(v) is not property:
         try:
            setattr(self, k, v)
         except:
            print('Cannot set attribute: {} to {}'.format(k, v))
            quit()
      self.data[k] = v

   def print(self):
      keyLen = 0
      for k in self.data.keys():
         keyLen = max(keyLen, len(k))

      print('Configuration')
      for k, v in self.data.items():
         print('   {:{}s}: {}'.format(k, keyLen, v))

class Config(Template):
   '''An environment configuration object

   Global constants are defined as static class variables. You can override
   any Config variable using standard CLI syntax (e.g. --NENT=128). 

   The default config as of v1.5 uses 1024x1024 maps with up to 2048 agents
   and 1024 NPCs. It is suitable to time horizons of 8192+ steps. For smaller
   experiments, consider the SmallMaps config.
   
   Notes:
      We use Google Fire internally to replace standard manual argparse
      definitions for each Config property. This means you can subclass
      Config to add new static attributes -- CLI definitions will be
      generated automatically.
   '''

   def __init__(self):
      super().__init__()

      if __debug__:
         err = 'config.Config is a base class. Use config.{Small, Medium Large}'''
         assert type(self) != Config, err

   ############################################################################
   ### Meta-Parameters
   RENDER                 = False
   '''Flag used by render mode'''

   def game_system_enabled(self, name) -> bool:
      return hasattr(self, name)

   ############################################################################
   ### Population Parameters                                                   
   AGENT_LOADER            = SequentialLoader
   '''Agent loader class specifying spawn sampling'''

   AGENTS = []
   '''Agent classes from which to spawn'''

   TASKS = []
   '''Tasks for which to compute rewards'''

   NMAPS                   = 1
   '''Number of maps to generate'''

   NTILE                   = 6
   #TODO: Find a way to auto-compute this
   '''Number of distinct terrain tile types'''

   NSTIM                   = 7
   '''Number of tiles an agent can see in any direction'''

   NMOB                    = None
   '''Maximum number of NPCs spawnable in the environment'''

   NENT                    = None
   '''Maximum number of agents spawnable in the environment'''

   NPOP                    = 1
   '''Number of distinct populations spawnable in the environment'''

   N_AGENT_OBS             = 100
   '''Number of distinct agent observations'''

   @property
   def TEAM_SIZE(self):
      assert not self.NENT % self.NPOP
      return self.NENT // self.NPOP

   @property
   def WINDOW(self):
      '''Size of the square tile crop visible to an agent'''
      return 2*self.NSTIM + 1

   ############################################################################
   ### Agent Parameters                                                   
   BASE_HEALTH                = 10
   '''Initial Constitution level and agent health'''

   PLAYER_SPAWN_ATTEMPTS      = None
   '''Number of player spawn attempts per tick

   Note that the env will attempt to spawn agents until success
   if the current population size is zero.'''

   def SPAWN_CONTINUOUS(self):
      '''Generates spawn positions for new agents

      Default behavior randomly selects a tile position
      along the borders of the square game map

      Returns:
         tuple(int, int):

         position:
            The position (row, col) to spawn the given agent
      '''
      #Spawn at edges
      mmax = self.TERRAIN_CENTER + self.TERRAIN_BORDER
      mmin = self.TERRAIN_BORDER

      var  = np.random.randint(mmin, mmax)
      fixed = np.random.choice([mmin, mmax])
      r, c = int(var), int(fixed)
      if np.random.rand() > 0.5:
          r, c = c, r 
      return (r, c)

   def SPAWN_CONCURRENT(self):
      left   = self.TERRAIN_BORDER
      right  = self.TERRAIN_CENTER + self.TERRAIN_BORDER
      rrange = np.arange(left+2, right, 4).tolist()

      assert not self.TERRAIN_CENTER % 4
      per_side = self.TERRAIN_CENTER // 4
      
      lows   = (left+np.zeros(per_side, dtype=np.int)).tolist()
      highs  = (right+np.zeros(per_side, dtype=np.int)).tolist()

      s1     = list(zip(rrange, lows))
      s2     = list(zip(lows, rrange))
      s3     = list(zip(rrange, highs))
      s4     = list(zip(highs, rrange))

      return s1 + s2 + s3 + s4

   @property
   def SPAWN(self):
      return self.SPAWN_CONTINUOUS

   ############################################################################
   ### Terrain Generation Parameters
   MAP_GENERATOR          = None
   '''Specifies a user map generator. Uses default generator if unspecified.'''

   FORCE_MAP_GENERATION   = False
   '''Whether to regenerate and overwrite existing maps'''

   GENERATE_MAP_PREVIEWS  = False
   '''Whether map generation should also save .png previews (slow + large file size)'''

   MAP_PREVIEW_DOWNSCALE  = 1
   '''Downscaling factor for png previews'''

   TERRAIN_CENTER             = None
   '''Size of each map (number of tiles along each side)'''

   TERRAIN_BORDER             = 16
   '''Number of lava border tiles surrounding each side of the map'''

   @property
   def TERRAIN_SIZE(self):
      return int(self.TERRAIN_CENTER + 2*self.TERRAIN_BORDER)

   TERRAIN_FLIP_SEED          = False
   '''Whether to negate the seed used for generation (useful for unique heldout maps)'''

   TERRAIN_FREQUENCY          = -3
   '''Base noise frequency range (log2 space)'''

   TERRAIN_FREQUENCY_OFFSET   = 7
   '''Noise frequency octave offset (log2 space)'''

   TERRAIN_LOG_INTERPOLATE_MIN = -2 
   '''Minimum interpolation log-strength for noise frequencies'''

   TERRAIN_LOG_INTERPOLATE_MAX= 0
   '''Maximum interpolation log-strength for noise frequencies'''

   TERRAIN_TILES_PER_OCTAVE   = 8
   '''Number of octaves sampled from log2 spaced TERRAIN_FREQUENCY range'''

   TERRAIN_LAVA               = 0.0
   '''Noise threshold for lava generation'''

   TERRAIN_WATER              = 0.30
   '''Noise threshold for water generation'''

   TERRAIN_GRASS              = 0.70
   '''Noise threshold for grass'''

   TERRAIN_FOREST             = 0.85
   '''Noise threshold for forest'''


   ############################################################################
   ### Path Parameters
   PATH_ROOT            = os.path.dirname(nmmo.__file__)
   '''Global repository directory'''

   PATH_CWD             = os.getcwd()
   '''Working directory'''

   PATH_RESOURCE        = os.path.join(PATH_ROOT, 'resource')
   '''Resource directory'''

   PATH_TILE            = os.path.join(PATH_RESOURCE, '{}.png')
   '''Tile path -- format me with tile name'''

   PATH_MAPS            = None
   '''Generated map directory'''

   PATH_MAP_SUFFIX      = 'map{}/map.npy'
   '''Map file name'''


############################################################################
### Game Systems (Static Mixins)
class Resource:
   '''Resource Game System'''

   @property #Reserved flag
   def Resource(self):
      return True

   RESOURCE_BASE_RESOURCE              = 10
   '''Initial level and capacity for Hunting + Fishing resource skills'''

   RESOURCE_FOREST_CAPACITY            = 1
   '''Maximum number of harvests before a forest tile decays'''

   RESOURCE_FOREST_RESPAWN             = 0.025
   '''Probability that a harvested forest tile will regenerate each tick'''

   RESOURCE_HARVEST_RESTORE_FRACTION   = 1.0
   '''Fraction of maximum capacity restored upon collecting a resource'''

   RESOURCE_HEALTH_REGEN_THRESHOLD     = 0.5
   '''Fraction of maximum resource capacity required to regen health'''

   RESOURCE_HEALTH_RESTORE_FRACTION    = 0.1
   '''Fraction of health restored per tick when above half food+water'''

class Combat:
   '''Combat Game System'''

   @property #Reserved flag
   def Combat(self):
      return True

   COMBAT_DICE_SIDES                   = 20
   '''Number of sides for combat dice

   Attacks can only hit opponents up to the attacker's level plus
   DICE_SIDES/2. Increasing this value makes attacks more accurate
   and allows lower level attackers to hit stronger opponents'''

   COMBAT_DEFENSE_WEIGHT               = 0.3 
   '''Fraction of defense that comes from the Defense skill'''   

   COMBAT_MELEE_REACH                  = 1
   '''Reach of attacks using the Melee skill'''

   COMBAT_RANGE_REACH                  = 3
   '''Reach of attacks using the Range skill'''

   COMBAT_MAGE_REACH                   = 4
   '''Reach of attacks using the Mage skill'''

   COMBAT_FREEZE_TIME                  = 3
   '''Number of ticks successful Mage attacks freeze a target'''


class Progression:
   '''Progression Game System'''

   @property #Reserved flag
   def Progression(self):
      return True

   PROGRESSION_BASE_RESOURCE           = 10
   '''Initial level and capacity for Hunting + Fishing resource skills'''

   PROGRESSION_BASE_XP_SCALE           = 10
   '''Skill level progression speed as a multiplier of typical MMOs'''

   PROGRESSION_CONSTITUTION_XP_SCALE   = 2
   '''Multiplier on top of XP_SCALE for the Constitution skill'''

   PROGRESSION_COMBAT_XP_SCALE         = 4
   '''Multiplier on top of XP_SCALE for Combat skills'''


class NPC(Combat):
   '''NPC & Equipment Game System'''

   @property #Reserved flag
   def NPC(self):
      return True

   NPC_SPAWN_ATTEMPTS                  = 25
   '''Number of NPC spawn attempts per tick'''

   NPC_SPAWN_AGGRESSIVE                = 0.80
   '''Percentage distance threshold from spawn for aggressive NPCs'''

   NPC_SPAWN_NEUTRAL                   = 0.50
   '''Percentage distance threshold from spawn for neutral NPCs'''

   NPC_SPAWN_PASSIVE                   = 0.00
   '''Percentage distance threshold from spawn for passive NPCs'''
   
   NPC_LEVEL_MIN                       = 1
   '''Minimum NPC level'''

   NPC_LEVEL_MAX                       = None
   '''Maximum NPC level'''

   NPC_LEVEL_SPREAD                    = None
   '''Level range for NPC spawns'''


class AllGameSystems(Resource, Progression, NPC): pass

############################################################################
### Config presets
class Small(Config):
   '''A small config for debugging and experiments with an expensive outer loop'''

   PATH_MAPS               = 'maps/small' 
   MAP_PREVIEW_DOWNSCALE   = 4

   TERRAIN_LOG_INTERPOLATE_MIN = 0

   TERRAIN_CENTER          = 32
   NENT                    = 64
   NMOB                    = 32

   PLAYER_SPAWN_ATTEMPTS   = 1

   NPC_LEVEL_MAX           = 10
   NPC_LEVEL_SPREAD        = 1

class Medium(Config):
   '''A medium config suitable for most academic-scale research'''

   PATH_MAPS               = 'maps/medium' 
   MAP_PREVIEW_DOWNSCALE   = 16

   TERRAIN_CENTER          = 128
   NENT                    = 256
   NMOB                    = 128

   PLAYER_SPAWN_ATTEMPTS   = 2

   NPC_LEVEL_MAX           = 30
   NPC_LEVEL_SPREAD        = 5

class Large(Config):
   '''A large config suitable for large-scale research or fast models'''

   PATH_MAPS               = 'maps/large' 
   MAP_PREVIEW_DOWNSCALE   = 64

   TERRAIN_CENTER          = 1024
   NENT                    = 2048
   NMOB                    = 1024

   PLAYER_SPAWN_ATTEMPTS   = 16

   NPC_LEVEL_MAX           = 99
   NPC_LEVEL_SPREAD        = 10

class Default(Medium, AllGameSystems): pass



from pdb import set_trace as T
from collections import defaultdict
from nmmo.lib import material
from copy import deepcopy
import os

import numpy as np
import json, pickle
import time
import ray

from nmmo.lib import utils

class Quill:
   def __init__(self):
      self.blobs = defaultdict(Blob)
      self.stats = defaultdict(list)

   def stat(self, key, val):
      self.stats[key].append(val)

   def register(self, key, tick, *plots):
      if key in self.blobs:
         blob = self.blobs[key]
      else:
         blob = Blob()
         self.blobs[key] = blob

      blob.tick = tick
      return blob

   @property
   def packet(self):
      logs = {key: blob.packet for key, blob in self.blobs.items()}
      return {'Log': logs, 'Stats': self.stats}

class Blob:
   def __init__(self):
      self.tracks = defaultdict(Track)
 
   def log(self, value, key=None):
      self.tracks[key].update(self.tick, value)

   @property
   def packet(self):
      return {key: track.packet for key, track in self.tracks.items()}

#Static blob analytics
class Track:
   def __init__(self):
      self.data = defaultdict(list)

   def update(self, tick, value):
      if type(value) != list:
         value = [value]

      self.data[tick] += value

   @property
   def packet(self):
      return self.data

#Log wrapper and benchmarker
class Benchmarker:
   def __init__(self, logdir):
      self.benchmarks = {}

   def wrap(self, func):
      self.benchmarks[func] = Utils.BenchmarkTimer()
      def wrapped(*args):
         self.benchmarks[func].startRecord()
         ret = func(*args)
         self.benchmarks[func].stopRecord()
         return ret
      return wrapped

   def bench(self, tick):
      if tick % 100 == 0:
         for k, benchmark in self.benchmarks.items():
            bench = benchmark.benchmark()
            print(k.__func__.__name__, 'Tick: ', tick,
                  ', Benchmark: ', bench, ', FPS: ', 1/bench)
 




from pdb import set_trace as T

from nmmo.lib.colors import Palette

class Agent:
    scripted = False
    name     = 'Neural_'

    def __init__(self, config, idx):
       '''Base class for agents

       Args:
          config: A Config object
          idx: Unique AgentID int
       '''
       self.config = config
       self.iden   = idx

    def __call__(self, obs):
       '''Used by scripted agents to compute actions. Override in subclasses.

       Args:
           obs: Agent observation provided by the environment
       '''
       pass

class Random(Agent):
    '''Moves randomly, including bumping into things and falling into lava'''
    def __call__(self, obs):
        return {Action.Move: {Action.Direction: rand.choice(Action.Direction.edges)}}



from pdb import set_trace as T

class Material:
   harvestable = False
   capacity    = 1
   def __init__(self, config):
      pass

   def __eq__(self, mtl):
      return self.index == mtl.index

   def __equals__(self, mtl):
      return self == mtl

class Lava(Material):
   tex   = 'lava'
   index = 0

class Water(Material):
   tex   = 'water'
   index = 1

class Grass(Material):
   tex   = 'grass'
   index = 2

class Scrub(Material):
   tex = 'scrub'
   index = 3

class Forest(Material):
   tex   = 'forest'
   index = 4

   harvestable = True
   degen       = Scrub

   def __init__(self, config):
      if config.game_system_enabled('Resource'):
         self.capacity = config.RESOURCE_FOREST_CAPACITY
         self.respawn  = config.RESOURCE_FOREST_RESPAWN

class Stone(Material):
   tex   = 'stone'
   index = 5

class Meta(type):
   def __init__(self, name, bases, dict):
      self.indices = {mtl.index for mtl in self.materials}

   def __iter__(self):
      yield from self.materials

   def __contains__(self, mtl):
      if isinstance(mtl, Material):
         mtl = type(mtl)
      if isinstance(mtl, type):
         return mtl in self.materials
      return mtl in self.indices

class All(metaclass=Meta):
   '''List of all materials'''
   materials = {Lava, Water, Grass, Scrub, Forest, Stone}

class Impassible(metaclass=Meta):
   '''Materials that agents cannot walk through'''
   materials = {Lava, Stone}

class Habitable(metaclass=Meta):
   '''Materials that agents cannot walk on'''
   materials = {Grass, Scrub, Forest}



from pdb import set_trace as T

from nmmo.systems.ai import behavior, utils

def passive(realm, entity):
   behavior.update(entity)
   actions = {}

   behavior.meander(realm, actions, entity)

   return actions

def neutral(realm, entity):
   behavior.update(entity)
   actions = {}

   if not entity.attacker:
      behavior.meander(realm, actions, entity)
   else:
      entity.target = entity.attacker
      behavior.hunt(realm, actions, entity)

   return actions

def hostile(realm, entity):
   behavior.update(entity)
   actions = {}

   # This is probably slow
   if not entity.target:
      entity.target = utils.closestTarget(entity, realm.map.tiles,
                                          rng=entity.vision)

   if not entity.target:
      behavior.meander(realm, actions, entity)
   else:
      behavior.hunt(realm, actions, entity)

   return actions



from pdb import set_trace as T
import numpy as np

from enum import Enum, auto

import nmmo
from nmmo.lib import utils
from nmmo.lib.utils import staticproperty

class NodeType(Enum):
   #Tree edges
   STATIC = auto()    #Traverses all edges without decisions
   SELECTION = auto() #Picks an edge to follow

   #Executable actions
   ACTION    = auto() #No arguments
   CONSTANT  = auto() #Constant argument
   VARIABLE  = auto() #Variable argument

class Node(metaclass=utils.IterableNameComparable):
   @staticproperty
   def edges():
      return []

   #Fill these in
   @staticproperty
   def priority():
      return None

   @staticproperty
   def type():
      return None

   @staticproperty
   def leaf():
      return False

   @classmethod
   def N(cls, config):
      return len(cls.edges)

   def args(stim, entity, config):
      return []

class Fixed:
   pass

#ActionRoot
class Action(Node):
   nodeType = NodeType.SELECTION

   @staticproperty
   def edges():
      '''List of valid actions'''
      return [Move, Attack]

   @staticproperty
   def n():
      return len(Action.arguments)

   def args(stim, entity, config):
      return nmmo.Serialized.edges 

   #Called upon module import (see bottom of file)
   #Sets up serialization domain
   def hook():
      idx = 0
      arguments = []
      for action in Action.edges:
         for args in action.edges:
            if not 'edges' in args.__dict__:
               continue
            for arg in args.edges: 
               arguments.append(arg)
               arg.serial = tuple([idx])
               arg.idx = idx 
               idx += 1
      Action.arguments = arguments

class Move(Node):
   priority = 1
   nodeType = NodeType.SELECTION
   def call(env, entity, direction):
      r, c  = entity.pos
      entID = entity.entID
      entity.history.lastPos = (r, c)
      rDelta, cDelta = direction.delta
      rNew, cNew = r+rDelta, c+cDelta
      
      #One agent per cell
      tile = env.map.tiles[rNew, cNew] 
      if tile.occupied and not tile.lava:
         return

      if entity.status.freeze > 0:
         return

      env.dataframe.move(nmmo.Serialized.Entity, entID, (r, c), (rNew, cNew))
      entity.base.r.update(rNew)
      entity.base.c.update(cNew)

      env.map.tiles[r, c].delEnt(entID)
      env.map.tiles[rNew, cNew].addEnt(entity)

      if env.map.tiles[rNew, cNew].lava:
         entity.receiveDamage(None, entity.resources.health.val)

   @staticproperty
   def edges():
      return [Direction]

   @staticproperty
   def leaf():
      return True

class Direction(Node):
   argType = Fixed

   @staticproperty
   def edges():
      return [North, South, East, West]

   def args(stim, entity, config):
      return Direction.edges

class North(Node):
   delta = (-1, 0)

class South(Node):
   delta = (1, 0)

class East(Node):
   delta = (0, 1)

class West(Node):
   delta = (0, -1)


class Attack(Node):
   priority = 0
   nodeType = NodeType.SELECTION
   @staticproperty
   def n():
      return 3

   @staticproperty
   def edges():
      return [Style, Target]

   @staticproperty
   def leaf():
      return True

   def inRange(entity, stim, config, N):
      R, C = stim.shape
      R, C = R//2, C//2

      rets = set([entity])
      for r in range(R-N, R+N+1):
         for c in range(C-N, C+N+1):
            for e in stim[r, c].ents.values():
               rets.add(e)

      rets = list(rets)
      return rets

   def l1(pos, cent):
      r, c = pos
      rCent, cCent = cent
      return abs(r - rCent) + abs(c - cCent)

   def call(env, entity, style, targ):
      if entity.isPlayer and not env.config.game_system_enabled('Combat'):
         return 

      #Check if self targeted
      if entity.entID == targ.entID:
         return

      #ADDED: POPULATION IMMUNITY
      #if entity.population == targ.population:
      #   return

      #Check attack range
      rng     = style.attackRange(env.config)
      start   = np.array(entity.base.pos)
      end     = np.array(targ.base.pos)
      dif     = np.max(np.abs(start - end))

      #Can't attack same cell or out of range
      if dif == 0 or dif > rng:
         return 
      
      #Execute attack
      entity.history.attack = {}
      entity.history.attack['target'] = targ.entID
      entity.history.attack['style'] = style.__name__
      targ.attacker = entity
      targ.attackerID.update(entity.entID)

      from nmmo.systems import combat
      dmg = combat.attack(entity, targ, style.skill)

      if style.freeze and dmg > 0:
         targ.status.freeze.update(env.config.COMBAT_FREEZE_TIME)

      return dmg

class Style(Node):
   argType = Fixed
   @staticproperty
   def edges():
      return [Melee, Range, Mage]

   def args(stim, entity, config):
      return Style.edges


class Target(Node):
   argType = None
   #argType = Player 

   @classmethod
   def N(cls, config):
      #return config.WINDOW ** 2
      return config.N_AGENT_OBS

   def args(stim, entity, config):
      #Should pass max range?
      return Attack.inRange(entity, stim, config, None)

class Melee(Node):
   nodeType = NodeType.ACTION
   index = 0
   freeze=False

   def attackRange(config):
      return config.COMBAT_MELEE_REACH

   def skill(entity):
      return entity.skills.melee

class Range(Node):
   nodeType = NodeType.ACTION
   index = 1
   freeze=False

   def attackRange(config):
      return config.COMBAT_RANGE_REACH

   def skill(entity):
      return entity.skills.range

class Mage(Node):
   nodeType = NodeType.ACTION
   index = 2
   freeze=True

   def attackRange(config):
      return config.COMBAT_MAGE_REACH

   def skill(entity):
      return entity.skills.mage

#TODO: Add communication
class Message:
   pass

#TODO: Add trade
class Exchange:
   pass

#TODO: Solve AGI
class BecomeSkynet:
   pass

Action.hook()



from pdb import set_trace as T
from itertools import chain

from setuptools import find_packages, setup


REPO_URL = "https://github.com/neuralmmo/environment"

extra = {
    'docs': [
        'sphinx-rtd-theme==0.5.1',
        'sphinxcontrib-youtube==1.0.1',
        ],
    'rllib': [
        'ray[all]==1.5.2',
        'tensorflow==2.7.0',
        'grpcio==1.32.0',
        'six==1.15.0',
        'dm-tree==0.1.6',
        'aiohttp==3.7',
        'pandas==1.3.4',
        'wandb==0.12.9',
        'openskill==0.2.0-alpha.0',
        'torch',
        ],
    }

extra['all'] = list(set(chain.from_iterable(extra.values())))
 
setup(
    name="nmmo",
    description="Neural MMO is a platform for multiagent intelligence research inspired by "
    "Massively Multiplayer Online (MMO) role-playing games. Documentation hosted at neuralmmo.github.io.",
    long_description_content_type="text/markdown",
    version=open('nmmo/version.py').read().split()[-1].strip("'"),
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'pytest-benchmark==3.4.1',
        'fire==0.4.0',
        'setproctitle==1.1.10',
        'service-identity==21.1.0',
        'autobahn==19.3.3',
        'Twisted==19.2.0',
        'vec-noise==1.1.4',
        'imageio==2.8.0',
        'tqdm==4.61.1',
        'pettingzoo==1.14.0',
    ],
    extras_require=extra,
    python_requires=">=3.9",
    license="MIT",
    author="Joseph Suarez",
    author_email="jsuarez@mit.edu",
    url=REPO_URL,
    keywords=["Neural MMO", "MMO"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
    ],
)




from pdb import set_trace as T
import numpy as np
from scipy import signal

def norm(ary, nStd=2):
   assert type(ary) == np.ndarray, 'ary must be of type np.ndarray'
   R, C         = ary.shape
   preprocessed = np.zeros_like(ary)
   nonzero      = ary[ary!= 0]
   mean         = np.mean(nonzero)
   std          = np.std(nonzero)
   if std == 0:
      std = 1
   for r in range(R):
     for c in range(C):
        val = ary[r, c]
        if val != 0:
           val = (val - mean) / (nStd * std)
           val = np.clip(val+1, 0, 2)/2
           preprocessed[r, c] = val
   return preprocessed

def clip(ary):
   assert type(ary) == np.ndarray, 'ary must be of type np.ndarray'
   R, C         = ary.shape
   preprocessed = np.zeros_like(ary)
   nonzero      = ary[ary!= 0]
   mmin         = np.min(nonzero)
   mmag         = np.max(nonzero) - mmin
   for r in range(R):
     for c in range(C):
        val = ary[r, c]
        val = (val - mmin) / mmag
        preprocessed[r, c] = val
   return preprocessed

def twoTone(ary, nStd=2, preprocess='norm', invert=False, periods=1):
   assert preprocess in 'norm clip none'.split()
   if preprocess == 'norm':
       ary   = norm(ary, nStd)
   elif preprocess == 'clip':
       ary   = clip(ary)

   R, C      = ary.shape

   colorized = np.zeros((R, C, 3))
   if periods != 1:
      ary = np.abs(signal.sawtooth(periods*3.14159*ary))
   if invert:
      colorized[:, :, 0] = ary
      colorized[:, :, 1] = 1-ary
   else:
      colorized[:, :, 0] = 1-ary
      colorized[:, :, 1] = ary

   colorized *= (ary != 0)[:, :, None]

   return colorized



from pdb import set_trace as T
import numpy as np
import random

from nmmo.lib.utils import inBounds
from nmmo.systems import combat
from nmmo.lib import material
from queue import PriorityQueue, Queue

from nmmo.systems.ai.dynamic_programming import map_to_rewards, \
   compute_values, max_value_direction_around

def validTarget(ent, targ, rng):
   if targ is None or not targ.alive:
      return False
   if lInf(ent, targ) > rng:
      return False
   return True


def validResource(ent, tile, rng):
   return tile is not None and tile.state.tex in (
      'forest', 'water') and distance(ent, tile) <= rng


def directionTowards(ent, targ):
   sr, sc = ent.base.pos
   tr, tc = targ.base.pos

   if abs(sc - tc) > abs(sr - tr):
      direction = (0, np.sign(tc - sc))
   else:
      direction = (np.sign(tr - sr), 0)

   return direction


def closestTarget(ent, tiles, rng=1):
   sr, sc = ent.base.pos
   for d in range(rng+1):
      for r in range(-d, d+1):
         for e in tiles[sr+r, sc-d].ents.values():
            if e is not ent and validTarget(ent, e, rng): return e

         for e in tiles[sr + r, sc + d].ents.values():
            if e is not ent and validTarget(ent, e, rng): return e

         for e in tiles[sr - d, sc + r].ents.values():
            if e is not ent and validTarget(ent, e, rng): return e

         for e in tiles[sr + d, sc + r].ents.values():
            if e is not ent and validTarget(ent, e, rng): return e

def distance(ent, targ):
   return l1(ent.pos, targ.pos)

def lInf(ent, targ):
   sr, sc = ent.pos
   gr, gc = targ.pos
   return abs(gr - sr) + abs(gc - sc)


def adjacentPos(pos):
   r, c = pos
   return [(r - 1, c), (r, c - 1), (r + 1, c), (r, c + 1)]


def cropTilesAround(position: (int, int), horizon: int, tiles):
   line, column = position

   return tiles[max(line - horizon, 0): min(line + horizon + 1, len(tiles)),
          max(column - horizon, 0): min(column + horizon + 1, len(tiles[0]))]


def inSight(dr, dc, vision):
    return (
          dr >= -vision and
          dc >= -vision and
          dr <= vision and
          dc <= vision)

def vacant(tile):
   from nmmo.io.stimulus.static import Stimulus
   Tile     = Stimulus.Tile
   occupied = Observation.attribute(tile, Tile.NEnts)
   matl     = Observation.attribute(tile, Tile.Index)

   lava    = material.Lava.index
   water   = material.Water.index
   grass   = material.Grass.index
   scrub   = material.Scrub.index
   forest  = material.Forest.index
   stone   = material.Stone.index
   orerock = material.Orerock.index

   return matl in (grass, scrub, forest) and not occupied

def meander(obs):
   from nmmo.io.stimulus.static import Stimulus

   agent  = obs.agent
   Entity = Stimulus.Entity
   Tile   = Stimulus.Tile

   r = Observation.attribute(agent, Entity.R)
   c = Observation.attribute(agent, Entity.C)

   cands = []
   if vacant(obs.tile(-1, 0)):
      cands.append((-1, 0))
   if vacant(obs.tile(1, 0)):
      cands.append((1, 0))
   if vacant(obs.tile(0, -1)):
      cands.append((0, -1))
   if vacant(obs.tile(0, 1)):
      cands.append((0, 1))
   if not cands:
      return (-1, 0)
   return random.choice(cands)

# A* Search
def l1(start, goal):
   sr, sc = start
   gr, gc = goal
   return abs(gr - sr) + abs(gc - sc)

def l2(start, goal):
   sr, sc = start
   gr, gc = goal
   return 0.5*((gr - sr)**2 + (gc - sc)**2)**0.5

#TODO: unify lInfty and lInf
def lInfty(start, goal):
   sr, sc = start
   gr, gc = goal
   return max(abs(gr - sr), abs(gc - sc))

def aStar(tiles, start, goal, cutoff=100):
   if start == goal:
      return (0, 0)

   pq = PriorityQueue()
   pq.put((0, start))

   backtrace = {}
   cost = {start: 0}

   closestPos = start
   closestHeuristic = l1(start, goal)
   closestCost = closestHeuristic

   while not pq.empty():
      # Use approximate solution if budget exhausted
      cutoff -= 1
      if cutoff <= 0:
         if goal not in backtrace:
            goal = closestPos
         break

      priority, cur = pq.get()

      if cur == goal:
         break

      for nxt in adjacentPos(cur):
         if not inBounds(*nxt, tiles.shape):
            continue
         if tiles[nxt].impassible:
            continue

         newCost = cost[cur] + 1
         if nxt not in cost or newCost < cost[nxt]:
            cost[nxt] = newCost
            heuristic = lInfty(goal, nxt)
            priority = newCost + heuristic

            # Compute approximate solution
            if heuristic < closestHeuristic or (
                    heuristic == closestHeuristic and priority < closestCost):
               closestPos = nxt
               closestHeuristic = heuristic
               closestCost = priority

            pq.put((priority, nxt))
            backtrace[nxt] = cur

   while goal in backtrace and backtrace[goal] != start:
      goal = backtrace[goal]

   sr, sc = start
   gr, gc = goal

   return (gr - sr, gc - sc)
# End A*

# Adjacency functions
def adjacentTiles(tiles, ent):
   r, c = ent.base.pos


def adjacentDeltas():
   return [(-1, 0), (1, 0), (0, 1), (0, -1)]


def l1Deltas(s):
   rets = []
   for r in range(-s, s + 1):
      for c in range(-s, s + 1):
         rets.append((r, c))
   return rets


def posSum(pos1, pos2):
   return pos1[0] + pos2[0], pos1[1] + pos2[1]


def adjacentEmptyPos(env, pos):
   return [p for p in adjacentPos(pos)
           if inBounds(*p, env.size)]


def adjacentTiles(env, pos):
   return [env.tiles[p] for p in adjacentPos(pos)
           if inBounds(*p, env.size)]


def adjacentMats(tiles, pos):
   return [type(tiles[p].state) for p in adjacentPos(pos)
           if inBounds(*p, tiles.shape)]


def adjacencyDelMatPairs(env, pos):
   return zip(adjacentDeltas(), adjacentMats(env.tiles, pos))
###End###



from pdb import set_trace as T

import scipy.stats as stats
import numpy as np

import os

import vec_noise
from imageio import imread, imsave
from tqdm import tqdm

from nmmo import material

def sharp(self, noise):
   '''Exponential noise sharpener for perlin ridges'''
   return 2 * (0.5 - abs(0.5 - noise));

class Save:
   '''Save utility for map files'''
   def render(mats, lookup, path):
      '''Render tiles to png'''
      images = [[lookup[e] for e in l] for l in mats]
      image = np.vstack([np.hstack(e) for e in images])
      imsave(path, image)

   def fractal(terrain, path):
      '''Render raw noise fractal to png'''
      frac = (256*terrain).astype(np.uint8)
      imsave(path, frac)

   def np(mats, path):
      '''Save map to .npy'''
      path = os.path.join(path, 'map.npy')
      np.save(path, mats.astype(int))

class Terrain:
   '''Terrain material class; populated at runtime'''
   pass

class MapGenerator:
   '''Procedural map generation'''
   def __init__(self, config):
      self.config = config
      self.loadTextures()

   def loadTextures(self):
      '''Called during setup; loads and resizes tile pngs'''
      lookup = {}
      path   = self.config.PATH_TILE
      scale  = self.config.MAP_PREVIEW_DOWNSCALE
      for mat in material.All:
         key = mat.tex
         tex = imread(path.format(key))
         lookup[mat.index] = tex[:, :, :3][::scale, ::scale]
         setattr(Terrain, key.upper(), mat.index)
      self.textures = lookup

   def generate_all_maps(self):
      '''Generates NMAPS maps according to generate_map

      Provides additional utilities for saving to .npy and rendering png previews'''

      config = self.config

      #Only generate if maps are not cached
      path_maps = os.path.join(config.PATH_CWD, config.PATH_MAPS)
      os.makedirs(path_maps, exist_ok=True)
      if not config.FORCE_MAP_GENERATION and os.listdir(path_maps):
          return

      if __debug__:
          print('Generating {} maps'.format(config.NMAPS))

      for idx in tqdm(range(config.NMAPS)):
         path = path_maps + '/map' + str(idx+1)
         os.makedirs(path, exist_ok=True)

         terrain, tiles = self.generate_map(idx)

         #Save/render
         Save.np(tiles, path)
         if config.GENERATE_MAP_PREVIEWS:
            b = config.TERRAIN_BORDER
            tiles = [e[b:-b+1] for e in tiles][b:-b+1]
            Save.fractal(terrain, path+'/fractal.png')
            Save.render(tiles, self.textures, path+'/map.png')

   def generate_map(self, idx):
      '''Generate a single map

      The default method is a relatively complex multiscale perlin noise method.
      This is not just standard multioctave noise -- we are seeding multioctave noise
      itself with perlin noise to create localized deviations in scale, plus additional
      biasing to decrease terrain frequency towards the center of the map

      We found that this creates more visually interesting terrain and more deviation in
      required planning horizon across different parts of the map. This is by no means a
      gold-standard: you are free to override this method and create customized terrain
      generation more suitable for your application. Simply pass MAP_GENERATOR=YourMapGenClass
      as a config argument.'''

      config      = self.config
      center      = config.TERRAIN_CENTER
      border      = config.TERRAIN_BORDER
      size        = config.TERRAIN_SIZE
      frequency   = config.TERRAIN_FREQUENCY
      offset      = config.TERRAIN_FREQUENCY_OFFSET
      octaves     = center // config.TERRAIN_TILES_PER_OCTAVE

      #Compute a unique seed based on map index
      #Flip seed used to ensure train/eval maps are different
      seed = idx + 1
      if config.TERRAIN_FLIP_SEED:
          seed = -seed

      #Log interpolation factor
      if not hasattr(self, 'interpolaters'):
         self.interpolaters = np.logspace(config.TERRAIN_LOG_INTERPOLATE_MIN,
               config.TERRAIN_LOG_INTERPOLATE_MAX, config.NMAPS)

      interpolate = self.interpolaters[idx]

      #Data buffers
      val   = np.zeros((size, size, octaves))
      scale = np.zeros((size, size, octaves))
      s     = np.arange(size)
      X, Y  = np.meshgrid(s, s)

      #Compute noise over logscaled octaves
      start = frequency
      end   = min(start, start - np.log2(center) + offset)
      for idx, freq in enumerate(np.logspace(start, end, octaves, base=2)):
         val[:, :, idx] = vec_noise.snoise2(seed*size + freq*X, idx*size + freq*Y)

      #Compute L1 distance
      x      = np.abs(np.arange(size) - size//2)
      X, Y   = np.meshgrid(x, x)
      data   = np.stack((X, Y), -1)
      l1     = np.max(abs(data), -1)

      #Interpolation Weights
      rrange = np.linspace(-1, 1, 2*octaves-1)
      pdf    = stats.norm.pdf(rrange, 0, interpolate)
      pdf    = pdf / max(pdf)
      high   = center / 2
      delta  = high / octaves

      #Compute perlin mask
      noise  = np.zeros((size, size))
      X, Y   = np.meshgrid(s, s)
      expand = int(np.log2(center)) - 2
      for idx, octave in enumerate(range(expand, 1, -1)):
         freq, mag = 1 / 2**octave, 1 / 2**idx
         noise    += mag * vec_noise.snoise2(seed*size + freq*X, idx*size + freq*Y) 

      noise -= np.min(noise)
      noise = octaves * noise / np.max(noise) - 1e-12
      noise = noise.astype(int)

      #Compute L1 and Perlin scale factor
      for i in range(octaves):
         start             = octaves - i - 1
         scale[l1 <= high] = np.arange(start, start + octaves)
         high             -= delta

      start   = noise - 1
      l1Scale = np.clip(l1, 0, size//2 - border - 2) 
      l1Scale = l1Scale / np.max(l1Scale)
      for i in range(octaves):
         idxs           = l1Scale*scale[:, :, i] + (1-l1Scale)*(start + i)
         scale[:, :, i] = pdf[idxs.astype(int)]

      #Blend octaves
      std = np.std(val)
      val = val / std
      val = scale * val
      val = np.sum(scale * val, -1)
      val = std * val / np.std(val)
      val = 0.5 + np.clip(val, -1, 1)/2

      #Threshold to materials
      matl = np.zeros((size, size), dtype=object)
      for y in range(size):
         for x in range(size):
            v = val[y, x]
            if v <= config.TERRAIN_WATER:
               mat = Terrain.WATER
            elif v <= config.TERRAIN_GRASS:
               mat = Terrain.GRASS
            elif v <= config.TERRAIN_FOREST:
               mat = Terrain.FOREST
            else:
               mat = Terrain.STONE
            matl[y, x] = mat

      #Lava and grass border
      matl[l1 > size/2 - border]       = Terrain.LAVA
      matl[l1 == size//2 - border]     = Terrain.GRASS

      edge  = l1 == size//2 - border - 1
      stone = (matl == Terrain.STONE) | (matl == Terrain.WATER)
      matl[edge & stone] = Terrain.FOREST

      return val, matl



from pdb import set_trace as T
import numpy as np

from collections import defaultdict
from collections.abc import Mapping
from typing import Dict, Callable

import nmmo
from nmmo import core, infrastructure
from nmmo.entity.npc import NPC
from nmmo.entity import Player
from nmmo.lib import colors

def prioritized(entities: Dict, merged: Dict):
   '''Sort actions into merged according to priority'''
   for idx, actions in entities.items():
      for atn, args in actions.items():
         merged[atn.priority].append((idx, (atn, args.values())))
   return merged

class EntityGroup(Mapping):
   def __init__(self, config, realm):
      self.dataframe = realm.dataframe
      self.config    = config

      self.entities  = {}
      self.dead      = {}

   def __len__(self):
      return len(self.entities)

   def __contains__(self, e):
      return e in self.entities

   def __getitem__(self, key):
      return self.entities[key]
   
   def __iter__(self):
      yield from self.entities

   def items(self):
      return self.entities.items()

   @property
   def corporeal(self):
      return {**self.entities, **self.dead}

   @property
   def packet(self):
      return {k: v.packet() for k, v in self.corporeal.items()}

   def reset(self):
      for entID, ent in self.entities.items():
         self.dataframe.remove(nmmo.Serialized.Entity, entID, ent.pos)

      self.spawned  = False
      self.entities = {}
      self.dead     = {}

   def add(iden, entity):
      assert iden not in self.entities
      self.entities[iden] = entity

   def remove(iden):
      assert iden in self.entities
      del self.entities[iden] 

   def spawn(self, entity):
      pos, entID = entity.pos, entity.entID
      self.realm.map.tiles[pos].addEnt(entity)
      self.entities[entID] = entity
 
   def cull(self):
      self.dead = {}
      for entID in list(self.entities):
         player = self.entities[entID]
         if not player.alive:
            r, c  = player.base.pos
            entID = player.entID
            self.dead[entID] = player

            self.realm.map.tiles[r, c].delEnt(entID)
            del self.entities[entID]
            self.realm.dataframe.remove(nmmo.Serialized.Entity, entID, player.pos)

      return self.dead

   def update(self, actions):
      for entID, entity in self.entities.items():
         entity.update(self.realm, actions)

class NPCManager(EntityGroup):
   def __init__(self, config, realm):
      super().__init__(config, realm)
      self.realm   = realm

   def reset(self):
      super().reset()
      self.idx     = -1

   def spawn(self):
      if not self.config.game_system_enabled('NPC'):
         return

      for _ in range(self.config.NPC_SPAWN_ATTEMPTS):
         if len(self.entities) >= self.config.NMOB:
            break

         center = self.config.TERRAIN_CENTER
         border = self.config.TERRAIN_BORDER
         r, c   = np.random.randint(border, center+border, 2).tolist()
         if self.realm.map.tiles[r, c].occupied:
            continue

         npc = NPC.spawn(self.realm, (r, c), self.idx)
         if npc: 
            super().spawn(npc)
            self.idx -= 1

   def actions(self, realm):
      actions = {}
      for idx, entity in self.entities.items():
         actions[idx] = entity.decide(realm)
      return actions
       
class PlayerManager(EntityGroup):
   def __init__(self, config, realm):
      super().__init__(config, realm)

      self.loader  = config.AGENT_LOADER
      self.palette = colors.Palette()
      self.realm   = realm

   def reset(self):
      super().reset()
      self.agents  = self.loader(self.config)
      self.idx     = 1

   def spawnIndividual(self, r, c):
      pop, agent = next(self.agents)
      agent      = agent(self.config, self.idx)
      player     = Player(self.realm, (r, c), agent, self.palette.color(pop), pop)
      super().spawn(player)
      self.idx   += 1

   def spawn(self):
      if self.config.SPAWN == self.config.SPAWN_CONCURRENT:
         if self.spawned:
            return 

         self.spawned = True
         idx = 0
         for r, c in self.config.SPAWN():
            idx += 1
            assert not self.realm.map.tiles[r, c].occupied
            self.spawnIndividual(r, c)
         return
          
      #MMO-style spawning
      for _ in range(self.config.PLAYER_SPAWN_ATTEMPTS):
         if len(self.entities) >= self.config.NENT:
            break

         r, c   = self.config.SPAWN()
         if self.realm.map.tiles[r, c].occupied:
            continue

         self.spawnIndividual(r, c)

      while len(self.entities) == 0:
         self.spawn()

class Realm:
   '''Top-level world object'''
   def __init__(self, config):
      self.config   = config

      #Generate maps if they do not exist
      config.MAP_GENERATOR(config).generate_all_maps()

      #Load the world file
      self.dataframe = infrastructure.Dataframe(config)
      self.map       = core.Map(config, self)

      #Entity handlers
      self.players  = PlayerManager(config, self)
      self.npcs     = NPCManager(config, self)

   def reset(self, idx):
      '''Reset the environment and load the specified map

      Args:
         idx: Map index to load
      ''' 
      self.map.reset(self, idx)
      self.players.reset()
      self.npcs.reset()
      self.tick = 0
 
   def packet(self):
      '''Client packet'''
      return {'environment': self.map.repr,
              'resource': self.map.packet,
              'player': self.players.packet,
              'npc': self.npcs.packet}

   @property
   def population(self):
      '''Number of player agents'''
      return len(self.players.entities)

   def entity(self, entID):
      '''Get entity by ID'''
      if entID < 0:
         return self.npcs[entID]
      else:
         return self.players[entID]

   def step(self, actions):
      '''Run game logic for one tick
      
      Args:
         actions: Dict of agent actions
      '''
      #Prioritize actions
      npcActions = self.npcs.actions(self)
      merged     = defaultdict(list)
      prioritized(actions, merged)
      prioritized(npcActions, merged)

      #Update entities and perform actions
      self.players.update(actions)
      self.npcs.update(npcActions)

      #Execute actions
      for priority in sorted(merged):
         for entID, (atn, args) in merged[priority]:
            ent = self.entity(entID)
            atn.call(self, ent, *args)

      #Spawn new agent and cull dead ones
      #TODO: Place cull before spawn once PettingZoo API fixes respawn on same tick as death bug
      self.players.spawn()
      self.npcs.spawn()

      dead = self.players.cull()
      self.npcs.cull()

      #Update map
      self.map.step()
      self.tick += 1

      return dead



from pdb import set_trace as T
import numpy as np

from nmmo import core
from nmmo.lib import material

import os

class Map:
   '''Map object representing a list of tiles
   
   Also tracks a sparse list of tile updates
   '''
   def __init__(self, config, realm):
      self.config = config

      sz          = config.TERRAIN_SIZE
      self.tiles  = np.zeros((sz, sz), dtype=object)

      for r in range(sz):
         for c in range(sz):
            self.tiles[r, c] = core.Tile(config, realm, r, c)

   @property
   def packet(self):
       '''Packet of degenerate resource states'''
       missingResources = []
       for e in self.updateList:
           missingResources.append(e.pos)
       return missingResources

   @property
   def repr(self):
      '''Flat matrix of tile material indices'''
      return [[t.mat.index for t in row] for row in self.tiles]

   def reset(self, realm, idx):
      '''Reuse the current tile objects to load a new map'''
      config = self.config
      self.updateList = set()

      path_map_suffix = config.PATH_MAP_SUFFIX.format(idx)
      fPath = os.path.join(config.PATH_CWD, config.PATH_MAPS, path_map_suffix)

      try:
         map_file = np.load(fPath)
      except FileNotFoundError:
         print('Maps not found')
         raise

      materials = {mat.index: mat for mat in material.All}
      for r, row in enumerate(map_file):
         for c, idx in enumerate(row):
            mat  = materials[idx]
            tile = self.tiles[r, c]
            tile.reset(mat, config)

   def step(self):
      '''Evaluate updatable tiles'''
      for e in self.updateList.copy():
         if e.static:
            self.updateList.remove(e)
         e.step()

   def harvest(self, r, c):
      '''Called by actions that harvest a resource tile'''
      self.updateList.add(self.tiles[r, c])
      return self.tiles[r, c].harvest()
